package PureMessage::Policy::Message;

use strict;

use PureMessage::Policy::Header;
use ActiveState::MIME::Parser;
use PureMessage::MIME::Encoding qw(make_encoder);
use PureMessage::Policy::Events qw(event_passed);
use PureMessage::_sys;
use Fcntl;
use ActiveState::UTF8 qw(encode_rfc2047);
use PureMessage::UTF8::Config qw(%utf8_config);

sub SRC_BUF_MAX     () { 1048576 }
sub BODY_BUF_MAX    () {   65536 }
sub DEF_BLOCK_SIZE  () {    8192 }

sub MAX_TOTAL_PARTS () {    1024 }
sub MAX_PART_DEPTH ()  {      16 }

sub MAX_PART_HEADERS() {     512 }

sub BODY_NOTSTARTED () { 0 }
sub BODY_INPROG     () { 1 }
sub BODY_FINISHED   () { 2 }

sub new {
    my $class = shift;
    my %opts = @_;
    my $self = ref $class ? $class : bless \%opts, $class;

    # quarantine option has to be supplied if invoking constructor as "static"
    if ((not exists $self->{base}) and (not $self->{quarantine})) {
        my $msg = "required option 'quarantine' is missing";
        $self->{logger}->log("ERROR", $msg);
	die "$msg";
    }

    $self->{part_id} ||= 1;
    $self->{depth}   ||= 0;
    $self->{body_decode_status} ||= BODY_NOTSTARTED;
    $self->{src_part_len} ||= 0;
    $self->{src_part_offset} ||= 0;
    $self->{block_size} ||= DEF_BLOCK_SIZE();
    $self->{descend_sub_message} = 1 unless exists $self->{descend_sub_message};
    $self->{base} ||= {
                         b_modified => 0,
                         src_buf => '',
                         event => 'connect',
                         assert_event => '',
                         src_file => '',
                         src_file_pos => 0,
                         quarantine => delete $self->{quarantine},
                         logger => delete $self->{logger},
                         block_size => delete $self->{block_size},
                         descend_sub_message => delete $self->{descend_sub_message},
                         tmp_id => delete $self->{tmp_id},

                         # Default maximum values (these are overridden by the
                         # policy engine, but should always be set).
                         max_total_parts => MAX_TOTAL_PARTS,
                         max_part_depth => MAX_PART_DEPTH,
			 max_part_headers => MAX_PART_HEADERS,
                         # detect 'preamble' and 'epilogue' parts for the root part (the message) ($self->{part_id} eq '1')
                         prelogue => {
                                        is_multipart => 0, # is the root part a multipart entity ?
                                        preamble_len => 0,
                                        epilogue_len => 0,
                                        got_boundary_callback => 0, # did we get boundary callback from mimeparser for the root part?
                                     },
                         mp_err => 0,
                      };
    return $self;
}

#sub DESTROY {
#    my $self = shift;
#    $self->log('DEBUG',$self->part_id." destroyed");
#}

sub get_info {
    my ($self, $key) = @_;
    return $self->{info}{$key};
}
sub set_info {
    my ($self, $key, $val) = @_;
    $self->{info}{$key} = $val;
}

sub set_limit {
    my ($self, @stuff) = @_;
    while (@stuff) {
	my ($k, $v) = splice(@stuff, 0, 2);
	if ($k =~ /^max_/ and exists $self->{base}{$k}) {
	    $self->{base}{$k} = $v if (defined($v));
	}
	else {
	    die "set_limit: unknown limit $k";
	}
    }
}

sub free {
    my $self = shift;
    
    # free myself and any children, grandchildren, ...
    my @parts = $self;
    while (my $part = shift @parts) {
        if ($part->{part_children}) {
            push @parts, values %{$part->{part_children}};
            delete $part->{part_children};
        }
        # mime parser contains circular refs
        delete $part->{parser};
        
        # close all decoded body data fds and tempfiles
        $part->body_close() if UNIVERSAL::can($part, "body_close");
    }
    
    # if i am the top-level message, free some additional shared objects
    if ($self->part_id eq '1') {
        # logger can contain circular ref
        delete $self->{base}{logger};
        
        # close the source data filehandle, if open
        my $fh = delete $self->{base}{src_file_handle};
        _close($fh) if defined $fh;
    }
}

sub print {
    my $self = shift;
    $self->{base}{src_buf} .= $_[0];
    $self->{src_part_len} += length $_[0];
    length $self->{base}{src_buf} > SRC_BUF_MAX and $self->_src_flush;    
}

#
# Events
#

sub assert_event {
    my $self = shift;
    my $event = shift;
    return unless $event;
    if (event_passed($self->{base}{event}, $event)) {
           delete $self->{base}{assert_event}; 
    }
    else {
        $self->{base}{assert_event} = $event;
        die "ENGINE EVENT $event\n";
    }
}

sub event {
    my $self = shift;
    my $event = shift;
    return $self->{base}{event} unless $event;
    $self->{base}{event} = $event;
#    $self->log('DEBUG',"event($event)");
}

#
# Size data
#

sub message_size {
    my $self = shift;
    $self->assert_event('eom');
    return $self->{src_part_len};
}

sub header_size {
    my $self = shift;
    $self->{src_hdr_parsed} or $self->_parse_headers;
    return $self->{src_hdr_len};    
}

sub body_raw_size {
    my $self = shift;
    $self->assert_event('eom');
    $self->{src_hdr_parsed} or $self->_parse_headers;
    return $self->{src_part_len} - $self->{src_hdr_len};
}

#
# Message source access
#

sub message_raw {
    my $self = shift;
    my $offset = $_[1] || 0;
    my $len = $_[2] || 0;
    
    # this can potentially be used to read header data,
    # so eom is not necessary
    
    $self->assert_event('eoh');
    
    $_[0] = '';
    
    my $avail = $self->{src_part_len} - $offset;
    return 0 unless $avail;
    (!$len or $len > $avail) and $len = $avail;
    $offset += $self->{src_part_offset};
    
    if ($self->{base}{src_file}) {
        $self->_src_flush;
        my $fh;
        unless (exists $self->{base}{src_file_handle}) {
            $self->{base}{src_file_handle} =
                _open($self->{base}{src_file}->filename, O_RDONLY);
            $self->{base}{src_file_pos} = 0;
        }
        if ($offset != $self->{base}{src_file_pos}) {
            $self->{base}{src_file_pos} =
                _lseek($self->{base}{src_file_handle}, $offset, 0);
        }
        my $rc = _read($self->{base}{src_file_handle}, $_[0], $len);
        $self->{base}{src_file_pos} += $rc;
        return $rc;
    }
    return length ($_[0] = substr $self->{base}{src_buf}, $offset, $len);
}

sub body_raw {
    my $self = shift;
    my $offset = $_[1] || 0;
    my $len = $_[2] || 0;
    
    $self->assert_event('eom');
    return $self->message_raw($_[0], $offset + $self->{src_hdr_len}, $len);
}

sub filename {
    my $self = shift;
    $self->assert_event('eom');
    $self->_src_flush;
    $self->{base}{src_file}->close;
    return $self->{base}{src_file}->filename;
}

sub tempfile {
    my $self = shift;
    $self->assert_event('eom');
    $self->_src_flush;
    return $self->{base}{src_file}; 
}

#
# Header access
#

sub header {
    my $self = shift;
    my $key = shift;
    my $index = shift || 0;
    my $decode = shift;
    
    $self->{src_hdr_parsed} or $self->_parse_headers;
    unless ($key) {
        if ($self->{h_modified}) {
            return $self->{header}->as_string;
        }
        else {
            return $self->{src_hdrs};
        }
    }
        
    # return the actual header, if it exists

    if (wantarray) {
        my @found;
        foreach (0 .. $self->{header}->count_fields($key) - 1) {
            push @found,  ($self->{header}->find_field($key, $_, !$decode))[1];
        }
        return @found;
    }
    else {
        return ($self->{header}->find_field($key, $index, !$decode))[1];
    }
}

sub header_decoded {
    my ($self, $key, $index) = @_;
    return $self->header($key, $index , 1);
}

# return header field names in order
sub headers {
    my $self = shift;
    $self->{src_hdr_parsed} or $self->_parse_headers;
    my (@headers, %headers);
    foreach (0 .. $self->{header}->count_fields() - 1) {
	my $h = ($self->{header}->get_field($_))[0];
        push @headers, $h unless $headers{lc $h}++;
    }
    return @headers;
}

sub headers_ordered {
    my $self = shift;
    $self->{src_hdr_parsed} or $self->_parse_headers;
    return @{$self->{header}->field_names};
}

sub header_add {
    my $self = shift;
    my $key = shift;
    my $value = shift;
    
    $self->{src_hdr_parsed} or $self->_parse_headers;
    eval {
        $self->{header}->add_field($key, encode_rfc2047($value), -1);
    }; 
    
    if ($@) {
	$self->throw('maxheaders', 'Message exceeds maximum number of headers') 
            if ($@ =~ /^too many headers/);
	$self->throw($@);
    }
    $self->{h_modified} = 1;
    $self->{base}{b_modified} = 1 unless $self->{part_id} eq '1';
}

sub header_delete {
    my $self = shift;
    my $key = shift;
    my $index = shift || 0;
    
    $self->{src_hdr_parsed} or $self->_parse_headers;

    ## Delete all matched headers for a given key
    if($index == -1) {
	for(;;) {
	    $index = $self->{header}->find_field($key, 0);
	    last if $index == -1;
	    $self->{header}->delete_field($index);
	}
    }
    else {
	my $i = $self->{header}->find_field($key, $index);
	return if $i == -1;
	$self->{header}->delete_field($i);
    }

    $self->{h_modified} = 1;
    $self->{base}{b_modified} = 1 unless $self->{part_id} eq '1';
}

sub header_replace {
    my $self = shift;
    my $key = shift;
    my $value = shift;
    my $index = shift || 0;
    
    $self->{src_hdr_parsed} or $self->_parse_headers;

    ## Replace all matched headers for a given key
    eval {
	if($index == -1) {
	    ## It returns 1 for unfound key as well.
	    my $c = $self->{header}->count_fields($key) || 1;
	    if($utf8_config{encoding}{guess}{replace_header} eq 'raw') {
		$self->{header}->replace_field($key, $value, $_) for(0..--$c);
	    }
	    else {
		$value = encode_rfc2047($value);
		$self->{header}->replace_field($key, encode_rfc2047($value), $_) for(0..--$c);
	    }
	}
	else {
	    if($utf8_config{encoding}{guess}{replace_header} eq 'raw') {
		$self->{header}->replace_field($key, $value, $index);
	    }
	    else {
		$self->{header}->replace_field($key, encode_rfc2047($value), $index);
	    }
	}
    };
    
    if ($@) {
	$self->throw('maxheaders', 'Message exceeds maximum number of headers') 
            if ($@ =~ /^too many headers/);
	$self->throw($@);
    }

    $self->{h_modified} = 1;
    $self->{base}{b_modified} = 1 unless $self->{part_id} eq '1';
}

sub header_exists {
    my $self = shift;
    my $key = shift;
    
    $self->{src_hdr_parsed} or $self->_parse_headers;
    return $self->{header}->count_fields($key) > 0;
}

sub content_type {
    my $self = shift;
    
    $self->{src_hdr_parsed} or $self->_parse_headers;
    my $cthdr = ($self->{header}->find_field('Content-Type', -1))[1];
    my ($ct,%ct_param) = $self->_content_type($cthdr);
    unless (defined $ct) {
	$ct = "text/plain";
    }
    $ct_param{charset} ||= "US-ASCII" if $ct =~ m,^text/,;
    return wantarray ? ($ct,%ct_param) : $ct;   
}

sub content_transfer_encoding {
    my $self = shift;
    
    $self->{src_hdr_parsed} or $self->_parse_headers;
    my $te = lc (($self->{header}->find_field('Content-Transfer-Encoding', -1))[1]);
    if ($te) {
	# fix some common misspellings
	$te =~ s/\s//g;
	if ($te eq "7-bit" || $te eq "8-bit") {
	    $te =~ s/-//;
	}
	elsif ($te eq "us-ascii" || $te eq "usascii") {
	    $te = "8bit";
	}
    }
    else {
	$te = "8bit";
    }
    return $te;
}

#
# Subparts access
#

sub parts {
    my $self = shift;
    $self->{body_parts_scanned} or $self->_parse_scan_parts;
    # Part IDs are of the form 1.NN, so we do a numeric
    # sort on the IDs after stripping their first two characters
    my @ids = sort { substr($a,2) <=> substr($b,2) }
	      grep {
		  # XXX this is a temporary work around to cope with
		  # partially formed message parts
		  my $p = $self->{part_children}{$_};
		  defined $p
		      and UNIVERSAL::isa($p, "PureMessage::Policy::Message")
	      }
	      keys %{$self->{part_children}};
    return @ids;
}

sub part {
    my $self = shift;
    my @ids = @_;
    return unless @ids;
    $self->{body_parts_scanned} or $self->_parse_scan_parts;
    wantarray or return $self->{part_children}->{$ids[0]};
    return @{$self->{part_children}}{@ids};
}

sub part_add {
    my $self = shift;
    my %opts = @_;
    
    $self->{body_parts_scanned} or $self->_parse_scan_parts;
    ++$self->{part_num_children};

    # Support "prepend => 1"
    $opts{before} = "1.1" if $opts{prepend};

    my ($rel_id, $abs_id);
    if ($opts{before}) {
	(my $location = $opts{before}) =~ s/^1\.//;

	# Make room for the new guy at the top
	my $kids = $self->{part_children};
	for my $id (reverse ($location .. $self->{part_num_children} - 1)) {
	    my $old_rel = "1.$id";
	    my $new_rel = "1.${\($id+1)}";
	    my $new_abs = "$self->{part_id}.${\($id+1)}";
	    $kids->{$new_rel} = delete $kids->{$old_rel};
	    $kids->{$new_rel}{part_id} = $new_abs;
	}
	$rel_id = "1.$location";
	$abs_id = "$self->{part_id}.$location";
    }
    else {
	$rel_id = "1.$self->{part_num_children}";
	$abs_id = "$self->{part_id}.$self->{part_num_children}";
    }

    $self->{part_children}->{$rel_id} = $self->_new_part(
	part_id => $abs_id,
	depth => $self->{depth} + 1,
	new_part => 1,
	base => {
	    quarantine => $self->{base}{quarantine},
	    src_buf => '',
	    event => $self->{base}{event},
	    block_size => $self->{base}{block_size},
	},
    );
    $self->{part_children}->{$rel_id}->_parse_headers;

    $self->{base}{b_modified} = 1;
    return $self->{part_children}->{$rel_id};
}

sub part_replace {
    my $self = shift;
    my $rel_id = shift;
    
    return unless defined $rel_id;
    
    $self->{body_parts_scanned} or $self->_parse_scan_parts;
    
    my @idparts = split /\./, $rel_id;
    my $abs_id = $self->{part_id}.".$idparts[1]";
    $self->{part_children}->{$rel_id}->free;
    $self->{part_children}->{$rel_id} = $self->_new_part(
                                            part_id=>$abs_id,
					    depth => $self->{depth} + 1,
                                            new_part => 1,
                                            base => {
                                                quarantine => $self->{base}{quarantine},
                                                src_buf => '',
                                                event => $self->{base}{event},
                                            },
                                        );
    $self->{base}{b_modified} = 1;
    return $self->{part_children}->{$rel_id};
}

sub part_delete {
    my $self = shift;
    my $id = shift;
    $self->{body_parts_scanned} or $self->_parse_scan_parts;
    return unless exists $self->{part_children}->{$id};
    $self->{part_children}->{$id}->free;
    delete $self->{part_children}->{$id};
    $self->{base}{b_modified} = 1;
}

sub part_offset {
    my $self = shift;
    $self->assert_event('eom');
    return $self->{src_part_offset};
}

sub part_id {
    return shift->{part_id};
}

sub part_nested {
    my ($self, $id, $abs) = @_;

    my @p = split /\./, $id;

    # If $absolute, remove $self->part_id from the beginning of $part_id.
    # Return nothing if the initial part isn't found.
    if ($abs) {
	my @s = split /\./, $self->part_id;
	for my $s (@s) {
	    my $p = shift @p;
	    defined $p and $p eq $s or return;
	}
    }

    my $part = $self;
    while (@p) {
	my $p = shift @p;
	$part = $part->part("1.$p") or return;
    }
    $part;
}

sub enumerate_parts {
	my $self = shift;
    my $cb = shift;
    
    my @parts = ($self);
    while (my $part = shift @parts) {
        my @ids = $part->parts();
        unshift @parts, map($part->part($_), @ids);
        $cb->($part);
    }
}

sub make_mixed {
    my $self = shift;
    my %opts = @_;
    my $prologue = defined $opts{prologue} ? $opts{prologue} : '';

    my $ct = $self->content_type;
    return if $ct eq 'multipart/mixed';

    $self->_decode_body()
        unless $self->{body_decode_status} == BODY_FINISHED
        or $self->{b_modified};

    my $new_part_id = "1." . ($self->{part_num_children} + 1);

    # in the last subpart we put the original body
    my $child = $self->part_add();
    $child->{body} = $self->{body};
    $child->{b_modified} = 1;
    $child->{src_hdr_parsed} = 1;
    $child->{body_decode_status} = BODY_FINISHED;
    $child->{part_boundary} = $self->{part_boundary};
    foreach(qw(Content-Type Content-Disposition Content-Transfer-Encoding)) {
        $child->header_add($_, $self->header($_)) if $self->header_exists($_);
        $self->header_delete($_) unless $_ eq 'Content-Type';
    }
    $child->{part_children} = delete $self->{part_children};
    delete $child->{part_children}{ $new_part_id };
    $self->{part_children}  = { "1.1" => $child };

    # finish setting up multipart
    $self->header_replace('Content-Type', "multipart/mixed; boundary=\"pmx-boundary-$self->{part_id}\"");
    $self->{part_boundary} = "pmx-boundary-$self->{part_id}";
    $self->{body} = {len => length($prologue), buf => $prologue}; 
    $self->{b_modified} = 0;
    $self->{base}{b_modified} = 1;
    $self->{part_num_children} = 1;

    return ($child);
}

sub make_multipart {
    my $self = shift;
    my %opts = @_;
    my $prologue = defined $opts{prologue} ? $opts{prologue} : '';
    my $prepend = $opts{prepend} || 0;
    
    my $ct = $self->content_type;
    return if $ct =~ m,^(?:multipart/|message/rfc822),;

    $self->_decode_body()
        unless $self->{body_decode_status} == BODY_FINISHED
        or $self->{b_modified};
    
    $self->{part_num_children} = 0; 
    $self->{is_multipart} = 1;
    
    # add desired number of subparts
    my @subparts;
    for (1 .. $prepend + 1) {
            push @subparts, $self->part_add();
    }
    
    # in the last subpart we put the original body
    my $last_sub_part = $subparts[$#subparts];
    $last_sub_part->{body} = $self->{body};
    $last_sub_part->{b_modified} = 1;
    $last_sub_part->{src_hdr_parsed} = 1;
    $last_sub_part->{body_decode_status} = BODY_FINISHED;
    foreach(qw(Content-Type Content-Disposition Content-Transfer-Encoding)) {
        $last_sub_part->header_add($_, $self->header($_)) if $self->header_exists($_);
        $self->header_delete($_) unless $_ eq 'Content-Type';
    }

    # finish setting up multipart
    $self->header_replace('Content-Type', "multipart/mixed; boundary=\"pmx-boundary-$self->{part_id}\"");
    $self->{part_boundary} = "pmx-boundary-$self->{part_id}";
    $self->{body} = {len => length($prologue), buf => $prologue}; 
    $self->{b_modified} = 0;
    $self->{base}{b_modified} = 1;
    @subparts;
}

## Decoded body access
#

sub body_filename {
    my $self = shift;
    $self->assert_event('eom');
    $self->_decode_body
        unless $self->{body_decode_status} == BODY_FINISHED;
    my $body = $self->{body};
    $self->_body_flush($body);
    $self->body_close;
    return $body->{tmpfile}->filename;   
}

sub body_close {
    my $self = shift;
    my $body = $self->{body};
    return unless $body;
    $self->_body_flush($body) if $body->{tmpfile};
    if (exists $body->{fh_r}) {
        _close($body->{fh_r});
        delete $body->{fh_r};
    }
    if (exists $body->{fh_w}) {
        _close($body->{fh_w});
        delete $body->{fh_w};
    }
    if ($body->{tmpfile} and !$body->{tmpclosed}) {
        $body->{tmpclosed} = 1;
        $body->{tmpfile}->close;
        $body->{filename} = $body->{tmpfile}->filename;
    }
}

sub body {
    my $self = shift;
    my $offset = $_[1] || 0;
    my $len = $_[2] || 0;
    
    $self->assert_event('eom');
    
    $_[0] = '';
    
    $self->_decode_body($len == 0 ? 0 : $offset + $len)
        unless $self->{body_decode_status} == BODY_FINISHED
        or $self->{b_modified};

    my $body = $self->{body};
    if ($body->{tmpfile}) {
        # if a tmpfile exists we need to open a new fh
        # to read back from it.
        unless (exists $body->{fh_r}) {
            my $filename = $self->body_filename;
            $body->{fh_r} = _open($filename, O_RDONLY);
            # pick up changes to the body length (i.e. virus cleaning)
            $body->{len} = (stat $filename)[7];
            $body->{pos} = 0;
        }
        return 0 unless $body->{len} and $offset < $body->{len};
        (!$len or $len > $body->{len}) and $len = $body->{len};
        unless ($body->{pos} == $offset) {
            $body->{pos} = _lseek($body->{fh_r}, $offset, 0);
        }
        my $rc = _read($body->{fh_r}, $_[0], $len);
        $body->{pos} += $rc;
        return $rc;
    }
    return 0 unless $body->{len} and $offset < $body->{len};
    (!$len or $len > $body->{len}) and $len = $body->{len};
    return length ($_[0] = substr $body->{buf}, $offset, $len);
}

sub body_append {
    my $self = shift;
    
    $self->assert_event('eom');
    
    $self->_decode_body() unless $self->{body_decode_status} == BODY_FINISHED;
    
    my $body = $self->{body};
    
    $self->_body_print($body, $_[0]);
    
    # close open fds on the body
    $self->body_close;
    
    $self->{b_modified} = 1;
    $self->{base}{b_modified} = 1;
}

sub body_copy_from {
    my $self = shift;
    my $fh_or_txt = shift;

    $self->assert_event('eom');
    
    # make sure we have at least parsed the headers of the part before
    # replacing the body
    $self->_parse_headers unless $self->{src_hdr_parsed};
    
    # close the old body
    $self->body_close;
    
    my $body = {};
    my $buf;
    
    if (ref($fh_or_txt)) {
	my $blksz = $self->{base}{block_size};
	while (read $fh_or_txt, $buf, $blksz) {
	    $self->_body_print($body, $buf);
	}
    }
    else {
	$self->_body_print($body, $fh_or_txt);
    }
    
    $self->{body} = $body;
    $self->{b_modified} = 1;
    $self->{base}{b_modified} = 1;
    
    # prevent this new body from being clobbered
    $self->{body_decode_status} = BODY_FINISHED;
}

# you can mark a message body as modified but not as un-modified

sub set_modified {
    my $self = shift;
    $self->{b_modified} = 1;
    $self->{base}{b_modified} = 1;
}

#
# Rebuild message with modifications
#
sub implode {
    my $self = shift;
    my $out = shift;
    my %opts = @_;
    
    my $cb;
    my $cb_p;
    $cb_p = $opts{cb_param} if defined $opts{cb_param};
    if (ref($out) eq 'CODE') {
        # this is a callback
        $cb = $out;
    }
    else {
        # else assume its a filehandle
        $cb = sub {print $out @_};
    }
    unless (delete $opts{no_header}) { 
        my $header = $self->header || "\r\n";
        $cb->(defined $cb_p ? ($cb_p, $header) : $header);
    }
    my $blksz = $self->{base}{block_size};
    if ($self->{part_id} eq '1' and !$self->{base}{b_modified}) {
        my ($read,$buf,$offset);
        while ($read = $self->body_raw($buf, $offset, $blksz)) {
            $read ? $offset += $read : last;
            $cb->(defined $cb_p ? ($cb_p,$buf) : $buf) if $buf;
        }
        return;
    }
    
    if ($self->{b_modified}) {
        my $te = $self->content_transfer_encoding;
        my $encoder = make_encoder($te) || make_encoder('8bit');
        my ($bufin, $offset) = ('', 0);
        while (my $read = $self->body($bufin, $offset, $blksz)) {
	    $offset += $read;
	    my $bufout = &$encoder($bufin);
            $cb->(defined $cb_p ? ($cb_p,$bufout) : $bufout) if length $bufout;
        }
	my $bufout = &$encoder();  # flush
	$cb->(defined $cb_p ? ($cb_p,$bufout) : $bufout) if length $bufout;
    }
    elsif ($self->{part_num_children}) {
        # this is a multipart message.  print the prologue.
        $self->body(my $buf);
        $cb->(defined $cb_p ? ($cb_p,$buf) : $buf);
    }
    else {
        my ($read, $buf,$offset);
        while (my $read = $self->body_raw($buf, $offset, $blksz)) {
            $read ? $offset += $read : last;
            $cb->(defined $cb_p ? ($cb_p,$buf) : $buf) if $buf;
        } 
    }    
    
    if (exists $self->{part_children} and %{$self->{part_children}}) {
        my $boundary = $self->{part_boundary};
        my $is_message_rfc822 = ($self->{ct} eq 'message/rfc822');
        my $any_children = 0;
        foreach ($self->parts) {
            next unless defined $self->{part_children}->{$_};
            $any_children = 1;
            # don't print a boundary if i'm a message/rfc822
            unless ($is_message_rfc822) {
                my $data = "\r\n--$boundary\r\n";
                $cb->(defined $cb_p ? ($cb_p,$data) : $data);
            }
            $self->{part_children}->{$_}->implode($cb,%opts);
        }
        # don't print a boundary if i'm a message/rfc822
        unless ($is_message_rfc822 or not $any_children) {
            my $data = "\r\n--$boundary--\r\n";
            $cb->(defined $cb_p ? ($cb_p,$data) : $data);
        }
    }    
}

sub log {
    my $self = shift;
    if ($self->{base}{logger}) {
        $self->{base}{logger}->log(@_);
    }
    else {
        shift;
        print @_, "\n";
    }
}

#
# IO functions
#

sub _open {
    my $file = shift;
    my $rc = PureMessage::_sys::open($file, @_);
    die "$file: open() failed: $!" if $rc == -1;
    return $rc;
}

sub _close {
    die "close(fd $_[0]) failed: $!" unless PureMessage::_sys::close(@_);
}

sub _read {
    my $rc = PureMessage::_sys::read(@_);
    die "read(fd $_[0]) failed: $!" if $rc == -1;
    return $rc;
}

sub _write {
    my $rc = PureMessage::_sys::write(@_);
    die "write(fd $_[0]) failed: $!" if $rc == -1;
    return $rc;
}

sub _lseek {
    my $rc = PureMessage::_sys::lseek(@_);
    die "lseek(fd $_[0]) failed: $!" if $rc == -1;
    return $rc;
}

#
# Parsing functions
#

sub _parse_scan_parts {
    my $self = shift;
    $self->assert_event('eom');
    $self->_parse(scan_parts_only=>1);
    $self->{body_parts_scanned} = 1;
}

sub _decode_body {
    my $self = shift;
    my $max = shift;
    $self->_parse(max_decoded_bytes=>$max);
}

sub _parse_headers {
    my $self = shift;
    $self->assert_event('eoh');
    $self->_parse(headers_only=>1);
    eval {
	$self->{header} =
	    PureMessage::Policy::Header->new(
		    defined $self->{src_hdrs} ?  $self->{src_hdrs} : "",
		    eol => "\r\n",
		    max_headers => $self->{base}{max_part_headers},
		);
    }; 
    
    if ($@) {
	$self->throw('maxheaders', 'Message exceeds maximum number of headers') 
            if ($@ =~ /^too many headers/);
	$self->throw($@);
    }
            
    $self->{src_hdr_parsed} = 1;
}

sub _parse {
    my $self = shift;
    my %opts = @_;
    
    my $headers_only = $opts{headers_only};
    my $scan_parts_only = $opts{scan_parts_only};
    my $max_decoded_bytes = $opts{max_decoded_bytes};
    
    if ($self->{body_decode_status} == BODY_NOTSTARTED) {
        $self->_init_parser();
        $self->{body} = {};
        $self->{body_decode_status} = BODY_INPROG;
        $self->{body_decode_bytes} = 0;
    }

    if ($self->{body_decode_status} == BODY_INPROG) {
        $self->{parser}->descend_multipart(1);
        $self->{parser}->descend_sub_message($self->{base}{descend_sub_message});
        my $offset = $self->{body_decode_pos};
        my ($buf,$read, $lastchr);
        while (1) {
            # stop conditions - we can stop parsing when one of these conditions
            # are met - it just means that we have enough data to satisfy the
            # current method call.
            last if ($headers_only and $self->{src_hdrs})
                    or ($max_decoded_bytes and $self->{body_decode_bytes} >= $max_decoded_bytes)
                    or ($scan_parts_only and $self->{src_hdrs} and !$self->{is_multipart});
            
            $read = $self->message_raw($buf,
                                       $offset,
                                       $self->{base}{block_size} );
            if ($read) {
                $offset += $read;
            }
            else {
                # If two boundaries occur one after the other, with no blank
                # lines between them, the newline after the first one will
                # be treated as the boundary data of the second one, and won't
                # be part of the subpart.  We add it here so that the subpart
                # parses properly.
                $self->{parser}->parse("\n")
                    if $self->{is_multipart} and $lastchr and $lastchr ne "\n";
                $self->{parser}->eof;
                last;
            }
            $lastchr = substr $buf, -1, 1;
            $self->{parser}->parse($buf);
        }
        $self->{body_decode_pos} = $offset;
    }
}

sub _init_parser {
    my $self = shift;
    $self->{parser} ||= ActiveState::MIME::Parser->new;
    $self->{parser}->descend_multipart(1);
    $self->{parser}->descend_sub_message($self->{base}{descend_sub_message});
    
    $self->{parser}->handler("part_start" =>
        sub {
        my $part_id = $_[0]->part_id;
        $self->{base}->{prelogue}->{is_multipart} = 1 if ($part_id eq '1' and $self->{part_id} eq '1' and $_[2] =~ m#^multipart/#);

        my $is_multi = ($_[2] =~ m#^multipart/# or $_[2] eq 'message/rfc822');
        my %info = (
                    src_hdr_len => length $_[1],
                    src_hdrs => $_[1],
                    ct => $_[2],
                    is_multipart => $is_multi,
                    ct_param => $_[3],
                    te => $_[4],
                    part_boundary => $_[3]->{boundary},
                    );
        
        my $part_suf = substr $part_id, 1;
        if ($part_suf) {
            $self->{part_num_children}++;
            $self->{last_part} = $part_id;
            $self->{last_offset} = $self->{body_decode_bytes};
            $info{part_id} = $self->{part_id}.$part_suf;
	    $info{depth}   = $self->{depth} + 1;
            $info{src_part_offset} = $self->{last_offset} + $self->{src_part_offset};
            $self->{part_children}->{$_[0]->part_id} = $self->_new_part(%info);
            $self->{parser}->descend_multipart(0);
            $self->{parser}->descend_sub_message(0);
        }
        elsif ($self->{part_id} eq '1' or $self->{new_part}) {
            while (my ($k,$v) = each %info) {
                $self->{$k} = $v;
            }
            $self->{src_part_offset} = 0;
        }
        $self->{body_decode_bytes} += length $_[1];
    });
    $self->{parser}->handler("part_end"   => sub {
        my $part_id = $_[0]->part_id;

        if ($part_id eq '1') {
            $self->{body_decode_status} = BODY_FINISHED;
        }
        else {
            $self->{part_children}->{$part_id}->{src_part_len} =
                $self->{body_decode_bytes} - $self->{last_offset};
            $self->{last_part} = $part_id;
            $self->{last_offset} = $self->{body_decode_bytes};
        }
    });
    $self->{parser}->handler("data_raw"       =>
        sub {
            if ($_[0]->part_id eq '1' and $self->{part_id} eq '1' and $self->{base}->{prelogue}->{is_multipart} and $_[1] !~ m/^\s+$/) {
                my $raw_data = $_[1];
                $raw_data =~ s/^\s+//; #ltrim
                $raw_data =~ s/\s+$//; #rtrim
                if ($self->{base}->{prelogue}->{got_boundary_callback}) {
                    $self->{base}->{prelogue}->{epilogue_len} += length $raw_data;
                }
                else {
                    $self->{base}->{prelogue}->{preamble_len} += length $raw_data;
                }
            }

            if ($self->{is_multipart}) {
                $self->{body_decode_bytes} += length $_[1];
                if ($_[0]->part_id eq '1') {
                    $self->_body_print($self->{body}, $_[1]);
                }
            }
        });
    
    $self->{parser}->handler("data_decoded"       =>
        sub {

            unless ($self->{is_multipart}) {
                $self->{body_decode_bytes} += length $_[1];
                if ($_[0]->part_id eq '1') {
                    $self->_body_print($self->{body}, $_[1]);
                }
            }
        });
    
    $self->{parser}->handler("boundary"   =>
        sub {
            $self->{base}->{prelogue}->{got_boundary_callback} = 1 if ($_[0]->part_id eq '1' and $self->{part_id} eq '1');
            $self->{body_decode_bytes} += length $_[1];
        });

    $self->{parser}->handler("error"   =>
        sub {
            $self->{base}->{mp_err} = $_[1];
            my $logline = "MIME parsing error (message offset " . $_[0]->offset . ", error code " . $_[1] . ")";
            if ($_[1] == ActiveState::MIME::Parser::MP_ERR_INVALID_CT_PARAM_SPEC) {
                $logline .= ": Content-Type header has invalid attribute specification";
            }
            $self->log('DEBUG',$logline);
        });

}

sub throw {
    my ($self, $exc, $message) = @_;
    $self->{base}{exception} = $exc;
    $self->{base}{exception_part} = $self->{part_id};
    $self->log("ERROR", "Message Exception: $exc");
    $self->log("ERROR", $message) if ($message);
    die "MESSAGE EXCEPTION: $exc\n";
}

sub _new_part {
    my $self = shift;
    my %opts = @_;

    if ($self->{base}{max_total_parts} && 
	++$self->{base}{total_parts} > $self->{base}{max_total_parts})
    {
	$self->throw("maxparts", "Message exceeds maximum number of MIME parts ("
		   . $self->{base}{max_total_parts}
		   . ")");
    }
    if ($self->{base}{max_part_depth} && 
	$self->{depth} >= $self->{base}{max_part_depth})
    {
	$self->throw("maxdepth", "Message exceeds maximum depth of MIME parts ("
		   . $self->{base}{max_part_depth}
		   . ")");
    }

    return PureMessage::Policy::Message->new(
        base => $self->{base},
        body_decode_bytes => 0,
        %opts
    );
}

sub _body_flush {
    my $self = shift;
    my $body = shift;
    $body->{tmpfile} ||= $self->new_tempfile;
    
    # if the tempfile has been closed,
    # we need to open it up again to
    # decode more data.
    my $buf = $body->{buf};
    return unless defined $buf and length $buf;
    if ($body->{tmpclosed}) {
        return unless -f $body->{filename};
        unless (exists $body->{fh_w}) {
            $body->{fh_w} = _open($body->{filename}, O_WRONLY | O_APPEND);
        }
        _write($body->{fh_w}, $buf);
    }
    else {
        $body->{tmpfile}->print($buf);
    }
    delete $body->{buf};
}

sub _body_print {
    my $self = shift;
    my $body = shift;
    
    if (ref $_[0]) {
        my $offset = 0;
        my $buf;
        my $blksz = $self->{base}{block_size};
        while (my $read = read $_[0], $buf, $blksz) {
            $read ? $body->{len} += $read : last;
            $body->{buf} .= $buf;
            length $body->{buf} > BODY_BUF_MAX and $self->_body_flush($body);
        }
    }
    else {
        $body->{buf} .= $_[0];
        $body->{len} += length $_[0];
        length $body->{buf} > BODY_BUF_MAX and $self->_body_flush($body);
    }
}


sub _src_flush {
    my $self = shift;
    return unless exists $self->{base}{src_buf};
    $self->{base}{src_file} ||= $self->new_tempfile;
    $self->{base}{src_file}->print($self->{base}{src_buf});
    delete $self->{base}{src_buf};
}

sub new_tempfile {
    my $self = shift;
    my $id = defined($self->{base}{tmp_id}) ? "-$self->{base}{tmp_id}" : "";
    my $uniq = PureMessage::_sys::tid() . $id;
    return $self->{base}{quarantine}->new_tempfile($uniq);
}

sub _d {
    shift->{debug} and print @_;
}

sub _content_type {
    my $self = shift;
    my $val = shift || '';
    
    # Extract the content type.
    return unless $val =~ s/^\s*([^\s;]+)//;  # XXX spaces in the type spec
    my $ct = lc($1);
    return $ct unless wantarray;

    # Extract parameters
    $val =~ s/\r//g;
    my @params;
    while ($val =~ s/^\s*;\s*([^\s=;]+)//) {
	push(@params, lc($1));  # attribute name

	# a quoted value
	if ($val =~ s/^\s*=\s*\"([^\"\\]*(?:\\.[^\"\\]*)*)\"//) {
	    my $v = $1;
	    $v =~ s/\\(.)/$1/g;
	    push(@params, $v);
	# some unquoted value
	} elsif ($val =~ s/^\s*=\s*([^;\s]*)//) {
	    my $v = $1;
	    $v =~ s/\s+$//;
	    push(@params, $v);
	# no value, a lone token
	} else {
	    push(@params, undef);
	}
    }

    return $ct, @params;
}

1;

__END__

=head1 NAME

PureMessage::Policy::Message - Policy engine message representation

=head1 SYNOPSIS

 # Create a new message object
 my $msg = PureMessage::Policy::Message->new();
 
 my $msg = PureMessage::Policy::Message->new(
                     quarantine => PureMessage::Quarantine->new,
                     logger => PureMessage::Policy::Logger->new,
 );
 
 # Feed it some message data
 $msg->print($msg_src);
 
 # Set the event to "End of Message"
 $msg->event('eom');
 
 # Get the 'Subject' header
 my $subject = $msg->header('subject');
 
 # Set the 'Subject' header
 $msg->header('subject','this is the new subject');
 
 # Decode the body to a buffer
 my $buf;
 my len =  $msg->body($buf);
 
 # Get the list of sub-parts
 my @sub_part_ids = $msg->parts();
 
 # Retrieve a subpart - another message object
 my $sub_part = $msg->part($sub_part_ids[3]);
 
 # Shortcut to get the content type
 my ($ct,$ct_params) = $sub_part->content_type;
 
 # Replace a part
 my $replacement_part = $msg->part_replace($sub_part_ids[3]);
 
 # Modify the replacement part
 $replacement_part->header('Content-Type','application/msword');
 $replacement_part->header('Content-Transfer-Encoding','base64');
 $replacement_part->copy_body_from('my_agenda.doc');
 
 # Write out the source of the modified message
 my $msg_src_out = $msg->implode();

=head1 DESCRIPTION

PureMessage::Policy::Message implements the core of the PureMessage message
functionality. It is (currently) closely tied to the PureMessage Policy
Engine.

=head2 Public Methods

The message object exposes the following methods:

=over

=item new()

 $msg = PureMessage::Policy::Message->new(%opts);

Creates a new message. The C<%opts> argument may contain the following keys:

=over

=item quarantine

A PureMessage::Quarantine object. See L<PureMessage::Quarantine>.

=item logger

A PureMessage::Logger object. See L<PureMessage::Logger>.

=back

=item free()

 $msg->free;

Frees memory associated with the message. The message object contains
self-referencial data structures that defeat Perl's resource management
system, so calling this method is required.

=item print()

 $msg->print("CHUNK");

Appends MIME data to the message. You can call this method multiple times to
append more and more data to it.

=item event()

 $msg->event("EVENT_NAME");

Records in the message that the event has occurred. Events must be one of the
following strings:

=over

=item connect

=item envelope

=item eoh

=item body

=item eom

=item disconnect

=item abort

=back

Events are temporally ordered as given above; so the C<eoh> event occurs later
than the C<envelope> event. See L<PerlMx> for details.

=item assert_event()

 $msg->assert_event("EVENT_NAME");

Throws an exception if the specified event has not yet occurred. This is the
means by which the PureMessage Policy Engine propagates the PerlMx events to
the policy script. Each method and plugin asserts that a particular event has
occurred, throwing an exception if it has not. The policy engine traps the
event, waits for that PerlMx event to be reached, then continues execution.

Most accessor methods call C<assert_event()>; these are documented below.

=item message_size()

 $size = $msg->message_size;

Returns the size of the message, in bytes.

Asserts that the C<eom> event has occurred.

=item header_size()

 $size = $msg->header_size;

Returns the size of the RFC 822 header block, in bytes.

Asserts that the C<eoh> event has occurred.

=item body_raw_size()

 $size = $msg->body_raw_size;

Returns the size of the message's raw (non-decoded) body, in bytes.

Asserts that the C<eom> event has occurred.

=item message_raw()

 $size = $msg->message_raw($scalar, $offset, $length);

Saves the requested substring of the raw message data into C<$scalar>, and
returns the number of bytes stored in C<$scalar>.

C<$offset> and C<$length> default to C<0>, so you can omit them.

Asserts that the C<eoh> event has occurred. If the amount of data requested is
not available, returns whatever data is available.

=item body_raw()

 $size = $msg->body_raw($scalar, $offset, $length);

Has the same semantics as C<message_raw()>, but retrieves the raw body, not
the raw message.

Asserts that the C<eom> event has occurred.

=item filename()

 $file = $msg->filename;

Returns the source filename containing the raw message data. If the message
was not originally read from a file, this creates a temporary file and saves
the message to it. The file is guaranteed to be closed.

Asserts that the C<eom> event has occurred.

=item tempfile()

 $file = $msg->tempfile;

Similar to the C<filename()> method, but the temporary file is kept open.

Asserts that the C<eom> event has occurred.

=item header()

 $rfc822 = $msg->header;
 $hdr    = $msg->header($header);
 @hdr    = $msg->header($header);
 $hdr    = $msg->header($header, $index);

With no arguments, returns the entire RFC 822 header block. If a C<$header> is
specified, the first occurrence of the header is returned in scalar context,
or every occurrences of the header in list context. In scalar context, a
particular occurrences of a C<$header> can be retrieved using the C<$index>
parameter.

Asserts that the C<eoh> event has occurred.

=item header_decoded()

 $hdr = $msg->header_decoded($header);
 @hdr = $msg->header_decoded($header);
 $hdr = $msg->header_decoded($header, $index);

Has the same semantics as C<header()>, but decodes RFC 2047 entities.

Asserts that the C<eoh> event has occurred.

=item headers()

 @hdrs = $msg->headers;

Returns the names of the headers in the order they first appear in the
message. For example, a typical return value might be:

 qw(Received To From Reply-To)

Asserts that the C<eoh> event has occurred.

=item header_add()

 $msg->header_add($key, $value);

Adds the C<$key> header, setting its value to C<$value>. Headers are added
after the existing headers.

Asserts that the C<eoh> event has occurred.

=item header_delete()

 $msg->header_delete($key, $index);

Deletes the C<$index>th C<$key> header, or the first if C<$index> is not
provided.

Asserts that the C<eoh> event has occurred.

=item header_replace()

 $msg->header_replace($key, $value, $index);

Replaces the C<$index>th C<$key> header with C<$value>, or the first if
C<$index> is not provided.

Asserts that the C<eoh> event has occurred.

=item header_exists()

 $exists = $msg->header_exists($header);

Returns a true value if the specified header exists, otherwise false.

Asserts that the C<eoh> event has occurred.

=item content_type()

 ($ct, %ct_param) = $msg->content_type;
 $ct = $msg->content_type;

Returns the normalized content type of the message. Reads the C<Content-Type>
header; if this is missing, defaults to C<text/plain>. In list context, also
returns the parameters from the C<Content-Type> header.

Asserts that the C<eoh> event has occurred.

=item content_transfer_encoding()

 $te = $msg->content_transfer_encoding;

Returns the normalized content transfer encoding of the message. Reads the
C<Content-Transfer-Encoding> header if present; if this is missing, defaults
to C<8bit>.

Asserts that the C<eoh> event has occurred.

=item parts()

 @ids = $msg->parts;

Returns a sorted list of subpart IDs that can be passed to C<part()> to access
the individual subparts. Note that only the direct subparts are returned, not
their subparts in turn.

Asserts that the C<eom> event has occurred.

=item part()

 $part  = $msg->part($id);
 @parts = $msg->part(@ids);

Returns the subpart corresponding the C<$id>. Each subpart is itself a
C<PureMessage::Policy::Message> object.

Asserts that the C<eom> event has occurred.

=item part_add()

 $part = $msg->part_add;

Adds a new, empty part at the end of the message and returns it. This is a
C<PureMessage::Policy::Message> object, so data can be fed to it using
C<print()> or other methods.

Asserts that the C<eom> event has occurred.

=item part_replace()

 $part = $msg->part_replace($id);

Replaces the specified subpart with a new, empty part and returns it. This is
a C<PureMessage::Policy::Message> object, so data can be fed to it using
C<print()> or other methods.

Asserts that the C<eom> event has occurred.

=item part_delete()

 $msg->part_delete($id);

Deletes the specified part from the message.  Make sure you don't call any
methods on the deleted part object (say, if you acquired the reference before
you deleted it from the parent).  To do so could cause the object to be cleaned
up improperly, and could have other unexpected effects.

Asserts that the C<eom> event has occurred.

=item part_offset()

 $offset = $msg->part_offset;

Returns the offset into the root message's source buffer where this part
begins.

Asserts that the C<eom> event has occurred.

=item part_id()

 $abs_id = $msg->part_id;

Returns the absolute part id of this message. This is typically different than
the return value of the C<part()> method, suitable for passing to
C<part_nested()>.

Asserts that the C<eom> event has occurred.

=item part_nested()

 $part = $msg->part_nested($abs_id);

Returns the specified part, traversing multiple layers of subparts if
necessary.

Asserts that the C<eom> event has occurred.

=item enumerate_parts()

 $msg->enumerate_parts(\&callback);

Invokes the callback for every part of the message (including the root message
itself). The callback is passed the C<PureMessage::Policy::Message> object
representing each part.

Asserts that the C<eom> event has occurred.

=item body_filename()

 $file = $msg->body_filename;

Returns the source filename containing the raw body data. This creates a
temporary file and saves the message to it. The file is guaranteed to be
closed.

Asserts that the C<eom> event has occurred.

=item body_close()

 $msg->body_close;

Closes the temporary file created with C<body_filename()>, if it exists.

Asserts that the C<eom> event has occurred.

=item body()

 $size = $msg->body($scalar, $offset, $length);

Has the same semantics as C<body_raw()>, but retrieves the decoded body, not
the raw message.

Asserts that the C<eom> event has occurred.

=item body_append()

 $msg->body_append($scalar);

Appends C<$scalar> to the decoded body.

Asserts that the C<eom> event has occurred.

=item body_copy_from()

 $msg->body_copy_from(\*STDIN);
 $msg->body_copy_from("STRING");

Replaces the body content from a filehandle or a string.

Asserts that the C<eom> event has occurred.

=item set_modified()

 $msg->set_modified;

Tell the message that its body has been modified. Useful if the file returned
from C<body_filename()> has been modified on disk and you want the message to
pick up the changes.

=item implode()

 $msg->implode($out, %opts);

Rebuilds the message with all modifications. C<$out> can be a filehandle, in
which case the message is printed to it; or a callback, in which case it is
invoked repeatedly with chunks of data which (when concatenated) form the raw
message.

The C<%opts> can have the following keys:

=over

=item cb_param

A paramter which will be passed to the callback before any others. If omitted,
the parameter is not passed at all.

=item no_header

If specified, the header is not emitted. By default, the entire message is
omitted.

=back

=item log()

 $msg->log($severity, $text);

Forwards the message to the internal C<PureMessage::Logger> object (or
whatever was passed to the constructor).

=back

=cut
