package PureMessage::Policy::MessageObject;

use base 'PureMessage::Policy::Message';
use strict;

use PureMessage::Config qw(%config);
use PureMessage::MessageInfo;

sub READ_BUF_MAX () { 2048 }

##################
# Public methods #
##################

#
# General
#

sub new {
    my $class = shift;
    my %opts = @_;
    my $self = bless \%opts, $class;
    $self->{minfo} ||= PureMessage::MessageInfo->new;
    $self->{buffer_size} = $config{buffer_size};
    $self->{vars} ||= {};
    $self->SUPER::new;
    return $self;
}

# signal that this message cannot be processed, for some reason. (e.g. too big,
# cannot be parsed, etc.)

sub message_failure {
    my $self = shift;
    my $reason = shift;
    
    # this should be caught and interpreted by the MessageStream driver
    $self->_croak("ENGINE FAILURE $reason");
}

#
# Metadata
#

sub minfo_value {
    my $self = shift;
    if (@_ > 1) {
        $self->{rcpt_modified} = 1 if lc($_[0]) eq 'envelope-to';
        $self->{from_modified} = 1 if lc($_[0]) eq 'envelope-from';
    }
    return $self->{minfo}->value(@_);
}

# this now is the preferred way to access envelope recipients

sub envelope_recipients {
    my $self = shift;
    my @old = $self->{minfo}->value('Envelope-To');
    if (@_) {
        $self->{rcpt_modified} = 1;
        $self->{minfo}->value('Envelope-To', @_);
    }
    return wantarray ? @old : scalar @old ? $old[0] : undef;
}

sub envelope_from {
    my $self = shift;
    my $old = $self->{minfo}->value('Envelope-From');
    if (@_) {
        $self->{from_modified} = 1;
        $self->{minfo}->value('Envelope-From', @_);
    }
    return $old;
}

sub mapped_recipients {
    my $self = shift;
    my @old = $self->{minfo}->value('Mapped-Recipients');
    if (@_) {
        $self->{minfo}->value('Mapped-Recipients', @_);
    }
    return wantarray ? @old : scalar @old ? $old[0] : undef;
}

sub minfo {
    return shift->{minfo};
}

sub get_attr {
    my $self = shift;
    my $k = shift;
    return $self->{mark}->{$k};
}

sub set_attr {
    my $self = shift;
    my $k = shift;
    my $v = shift;
    my $is_contextual = shift;
    $self->{mark}->{$k} = $v;
    if ($is_contextual) {
        my $context = pop @{$self->{context}};
        push @$context, $k;
        push @{$self->{context}}, $context;
    }
}

sub del_attr {
    my $self = shift;
    my $k = shift;
    delete $self->{mark}->{$k};
}

sub attrs {
    my $self = shift;
    return keys %{$self->{mark}};
}

sub context_push {
    my $self = shift;
    push @{$self->{context}}, [];
}

sub context_pop {
    my $self = shift;
    my $context = pop @{$self->{context}};
    foreach my $attr (@$context) {
        pop @{$self->{mark}->{$attr}};
    }
}

sub context_reset {
    my $self = shift;
    delete $self->{context};
    delete $self->{context_flip};
}

sub context_flip {
    my $self = shift;
    my $flip = shift;
    $self->{context_flip} = (not $self->{context_flip}) if $flip;
    return $self->{context_flip};
}

sub context_not {
    my $self = shift;
    my $old = $self->{context_flip};
    $self->{context_flip} = shift if @_;
    return $old;
}

#
# Delivery
#

sub delivery_action {
    my $self = shift;
    my $action = shift;
    my $reason = shift;
    my $rcode = shift;
    my $xcode = shift;
    return $self->{delivery_action} unless $action;
    
    # the delivery_action is read-only after it is first set
    $self->{delivery_action} = {
                                action => $action,
                                reason => $reason,
                                rcode => $rcode || 550,
                                xcode => $xcode,
                               }
        unless $self->{delivery_action};   
}

sub stop {
    my $self = shift;
    $self->{stopped} = 1;
    return 1;
}

sub stopped {
    my $self = shift;
    return $self->{stopped};
}

sub file {
    my $self = shift;
    my $reason = shift;
    my %opts = @_;

    my $tmp;    
    if ($opts{discard} and not($self->{base}{b_modified} or $self->{h_modified})) {
        $tmp = $self->tempfile;
    }
    else {
        $self->fixup_message_id();
        $tmp = $self->implode_to_tempfile();
    }
    
    # ensure reason sanity - no empty strings
    $reason ||= "None";
    
    # ensure reason sanity - no whitespace
    $reason =~ s/\s+/_/g;
    
    $self->minfo_value("Reason",$reason);
    $tmp->store($self->{minfo});
}

sub implode_to_tempfile {
    my $self = shift;
    my $tmp = $self->new_tempfile;
    my $tmpfile = $tmp->filename;
    open my $fh,  ">$tmpfile" or die "$tmpfile: $!";
    $self->implode($fh);
    close $fh;
    $tmp;
}

sub tmp_id {
    my $self = shift;
    my $id = shift;
    return $self->{base}{tmp_id} unless defined $id;
    $self->{base}{tmp_id} = $id;
}

sub modified {
    my $self= shift;
    return ($self->{base}{b_modified} or $self->{h_modified} or $self->{rcpt_modified} or $self->{from_modified});
}

sub body_modified {
    return shift->{base}{b_modified};
}

sub header_modified {
    return shift->{h_modified};
}

sub rcpt_modified {
    return shift->{rcpt_modified};
}

sub from_modified {
    return shift->{from_modified};
}

sub get_header_additions {
    return shift->{header}->get_additions;
}

sub get_header_replacements {
    return shift->{header}->get_replacements;
}

sub fixup_message_id {
    my $self = shift;

    # Only modifiy the message-id if we are allowed to
    return if $config{preserve_msgid};

    if ( $self->body_modified() or 
           ( (not $config{'preserve_msgid_hdr_change'}) and $self->header_modified() and $self->get_header_replacements() )
       ) {
        my $msgid = $self->header('Message-ID');

        # The MTA should be responsible for adding a Message-ID.
        if ($msgid) {
            my $tm = time();
            my $tmpid = $self->tmp_id() || $tm;
            $msgid  = $1 if $msgid =~ /<(.*)>/;
            my $newid = join('_', $$, $tm, $tmpid, $msgid);
            $self->header_replace('Message-ID', "<$newid>");
        }
    }
}

sub dkim_verify {
    my $self = shift;
    eval {    
        require Mail::DKIM::Verifier;
    };
    if ($@) {
        $self->log('ERROR',"Mail::DKIM has not installed: '$@'");
        return 'invalid';
    }

    my $dkim;
    eval {
        $dkim = Mail::DKIM::Verifier->new();
        my $msgfile = $self->filename();
        open my $fh, $msgfile or die "$msgfile: $!";
        while (<$fh>) {
            # remove local line terminators
            chomp;
            s/\015$//;
            # use SMTP line terminators
            $dkim->PRINT("$_\015\012");
        }
        $dkim->CLOSE;
        close $fh;
    };
    if ($@) {
        $self->log('ERROR',"DKIM-VERIFY: Error '$@'");
        return 'invalid'; # unknown error, dkim dies
    }

    my $result = $dkim->result();
    if ($result eq 'fail' or $result eq 'invalid') {
        my $detail = $dkim->result_detail() || '';
        $self->log('ERROR',"DKIM-VERIFY: '$result' and detail info: '$detail'");
    }
    return $result;
}

sub dkim_sign {
    my $self = shift;
    
    my $algorithm;
    my $method;
    my $domain;
    my $selector;
    my $keyfile;

    eval {
        require Mail::DKIM::Signer;
        require PureMessage::Policy::DKIM::Config;

        my $config = \%PureMessage::Policy::DKIM::Config::dkim_config;
        $algorithm = $config->{algorithm};
        $method    = $config->{method};
        $domain    = $config->{domain};
        $selector  = $config->{selector};
        $keyfile   = $config->{keyfile};

        for my $item (qw(algorithm method domain selector keyfile)) {
            my $value = $config->{$item};
            unless ($value) {
                $self->set_attr("mark_DKIM-SIGN", "no_$item");
                $self->log('ERROR',"DKIM-SIGN: '$item' not defined");
                return;
            }
            if ($item eq 'keyfile' and (!-f $value)) {
                $self->set_attr("mark_DKIM-SIGN", "keyfile_notfound");
                $self->log('ERROR',"DKIM-SIGN: key file '$value' not found");
                return;
            }
        }
    };
    if ($@) {
        $self->log('ERROR',"Mail::DKIM has not installed: '$@'");
        return;
    }

    my $dkim; 
    my $dkim_sig;
    eval { 
        $dkim = Mail::DKIM::Signer->new(
                      Algorithm => $algorithm,
                      Method => $method,
                      Domain => $domain,
                      Selector => $selector,
                      KeyFile => $keyfile,
        );

        my $tmp     = $self->implode_to_tempfile();
        my $msgfile = $tmp->filename();
        open my $fh, $msgfile or die "$msgfile: $!";
        while (<$fh>) {
            # remove local line terminators
            chomp;
            s/\015$//;

            # use SMTP line terminators
            $dkim->PRINT("$_\015\012");
        }
        $dkim->CLOSE;
        close $fh;

        # what is the signature result?
        my $signature = $dkim->signature;
        $dkim_sig = $signature->as_string;
    };
    if ($@) {
        $self->log('ERROR',"DKIM-SIGN: Error '$@'");
        $self->set_attr("mark_DKIM-SIGN", "error");
        return;
    }
    my ($dkim_key, $dkim_val) = split /:/, $dkim_sig, 2;
    $dkim_val =~ s/^\s+//;

    $self->header_add($dkim_key, $dkim_val) if $dkim_key and $dkim_val;
    $self->log('DEBUG',"DKIM-SIGN: Signature: '$dkim_val'");
    $self->set_attr("mark_DKIM-SIGN", "ok");
}

sub template {
    return shift->{template}->template(@_);
}

sub template_expand {
    return shift->{template}->expand(@_);
}

sub template_object {
    my $self = shift;
    my $o = shift;
    
    return $self->{template} unless defined $o;
    $self->{template} = $o;
}

sub clone {
    my $self = shift;
    my %minfo_items = @_;
    
    $self->assert_event('eom');
    
    my $class = ref $self;
    
    my $clone = $class->new(
        minfo => PureMessage::MessageInfo->new($self->{minfo}->items()),
        logger => $self->{base}{logger},
        quarantine => $self->{base}{quarantine},
        tmp_id => $self->{base}{tmp_id},
    );
    
    # set minfo items
    foreach my $key (keys %minfo_items) {
        my $value = $minfo_items{$key};
        my @values = ref $value ? @$value : $value;
        $clone->minfo->value($key, @values);
    }
    
    # XXX copy message source.  this sucks - we should find a way to share
    # XXX these parts between message objects.
    my ($offset, $block_size, $count, $buf) = (0, $self->{base}{block_size});
    while ($count = $self->message_raw($buf, $offset, $block_size)) {
        $clone->print($buf);
        $offset += $count;
    }
    $clone->event($self->{base}{event});
    
    return $clone;
}

1;
